package App::Greple::xlate::null;

our $VERSION = "1.0202";

use v5.14;
use warnings;
use Encode;
use Data::Dumper;

use App::Greple::xlate qw(opt);
use App::Greple::xlate::Lang qw(%LANGNAME);

sub xlate {
    @_;
}

1;

__DATA__

option default -Mxlate --xlate-engine=null
